
import six
import sys
import os
import io
from say import *
import pytest

globalvar = 99


def test_basic(param='Yo'):
    say = Fmt()

    greeting = "hello"
    assert say("{greeting}, world!") == "{0}, world!".format(greeting)
    assert say("'{greeting}' has {len(greeting)} characters") == "'{0}' has {1} characters".format(greeting, len(greeting))
    assert say("{param}") == "{0}".format(param)
    assert say("{greeting:<10}") == "{0:<10}".format(greeting)
    assert say("{greeting:>20}") == "{0:>20}".format(greeting)
    assert say("{globalvar:3}") == " 99"
    assert say("{globalvar:4}") == "  99"
    assert say("{globalvar:<4}") == "99  "

    status = "OK"
    assert say("This is {status}", silent=True) == "This is OK"


def test_gt():
    say = Fmt()

    x = 555
    assert (say > "{x} is a big number!") == "555 is a big number!"

def test_localvar():
    say = Fmt()

    m = "MIGHTY"
    assert say(m) == "MIGHTY"
    assert say("{m}") == "MIGHTY"
    globalvar = "tasty"    # override global var
    assert say(globalvar) == "tasty"
    assert say("{globalvar}") == "tasty"

def test_globalvar():
    say = Fmt()

    assert say("{globalvar}") == str(globalvar)

def test_unicode():
    say = Fmt()

    u = six.u('This\u2014is Unicode!')
    assert say(u) == u
    assert say("Hey! {u}") == six.u('Hey! This\u2014is Unicode!')
    too = "too"
    assert say(six.u("Unicode templates {too}")) == six.u("Unicode templates too")

def test_format_string():
    say = Fmt()

    x = 33.123456789
    assert say("{x} is floating point!") == '33.123456789 is floating point!'
    assert say("{x:0.2f} is shorter") == '33.12 is shorter'
    assert say("{x:8.3f} is in the middle") == '  33.123 is in the middle'


def test_files(capsys, tmpdir):
    say = Say()

    tmpfile = tmpdir.join('test.txt')
    say.setfiles([sys.stdout, tmpfile])

    text = "Yowza!"
    say(text)

    assert tmpfile.read() == text + "\n"
    assert capsys.readouterr()[0] == text + "\n"

    text2 = six.u('Hello\u2012there')
    tmpfile2 = tmpdir.join('test2.txt')

    tfname = tmpfile2.strpath
    say.set(files=[sys.stdout, tfname])
    say(text2)
    tf = say.options.files[1]
    tf.close()
    with io.open(tfname, 'r') as tf2:
        assert tf2.read() == text2 + "\n"
    assert capsys.readouterr()[0] == text2 + "\n"

    errfile = tmpdir.join('error.txt')
    errcode = 12
    err = say.clone(files=[sys.stderr, errfile])
    err("Failed with error {errcode}")  # writes to stderr, error.txt
    assert capsys.readouterr()[1] == 'Failed with error 12\n'
    assert errfile.read()         == 'Failed with error 12\n'


def test_example_1():
    say = Fmt()

    x = 12
    nums = list(range(4))
    name = 'Fred'

    assert say("There are {x} things.") == "There are 12 things."
    assert say("Nums has {len(nums)} items: {nums}") == "Nums has 4 items: [0, 1, 2, 3]"
    assert say("Name: {name!r}") == "Name: 'Fred'"

    assert say("There are {x} things.") == "There are {0} things.".format(x)
    assert say("Nums has {len(nums)} items: {nums}") == "Nums has {0} items: {1}".format(len(nums), nums)
    assert say("Name: {name!r}") == "Name: {0!r}".format(name)


def test_example_2():
    say = Fmt()

    errors = [{'name': 'I/O Error', 'timestamp': 23489273},
              {'name': 'Compute Error', 'timestamp': 349734289}]

    say.title('Errors')
    for i, e in enumerate(errors, start=1):
        say("{i:3}: {e['name'].upper()}")


def test_say_silent():
    say = Fmt()
    r1 = say("this is output")
    r2 = say("this is output", silent=True)
    assert r1 == r2


def test_indent():
    say = Fmt()
    assert say('no indent') == 'no indent'
    assert say('no indent', indent=0) == 'no indent'
    assert say('one indent', indent='1') == '    one indent'
    assert say('one indent', indent='+1') == '    one indent'
    assert say('two indent', indent=2) == '        two indent'
    say.set(indent=1)
    assert say('auto one indent') == '    auto one indent'
    assert say('one plus one indent', indent='+1') == '        one plus one indent'
    assert say('subtract indent', indent='-1') == 'subtract indent'
    assert say('force no indent', indent=0) == 'force no indent'
    say.set(indent=0)
    assert say('no indent again') == 'no indent again'


def test_sep_end():
    say = Fmt()
    assert say(1, 2, 3) == '1 2 3'
    assert say(1, 2, 3, sep=',') == '1,2,3'
    assert say(1, 2, 3, sep=', ') == '1, 2, 3'
    say('xyz', end='\n') == 'xyz\n'
    say('xyz', end='\n\n') == 'xyz\n\n'
    say('xyz', end='') == 'xyz'


def test_indent_special():
    say = Fmt()
    say.set(indent_str='>>> ')
    assert say('something') == 'something'
    assert say('else', indent=1) == '>>> else'
    assert say('again', indent=2) == '>>> >>> again'
    say.set(indent_str='| ')
    assert say("some text") == 'some text'
    assert say("other", indent="+1") == '| other'


def test_indent_multiline():
    say = Fmt()
    assert say('and off\nwe go', indent='+1') == '    and off\n    we go'


def test_with_indent():
    say = Fmt()
    with say.settings(indent='+1'):
        assert say("I am indented!") == "    I am indented!"
        with say.settings(indent='+1'):
            assert say("xyz") == "        xyz"
        assert say('one back') == '    one back'
    assert say('back again') == 'back again'

def test_prefix_suffix():
    say = Fmt()
    assert say('x', prefix='<', suffix='>') == '<x>'
    assert say('x', end='\n--') == 'x\n--'
    assert say('a\nb\nc', prefix='> ') == '> a\n> b\n> c'
    assert say('a\nb\nc', prefix='> ', suffix='\n') == '> a\n\n> b\n\n> c\n'
    quoter = say.clone(prefix='> ')
    assert quoter('a\nb\nc') == '> a\n> b\n> c'

def test_prefix_in_context_example():
    say = Fmt()
    with say.settings(prefix='> '):
        assert say('this') == '> this'
        assert say('that') == '> that'
    assert say('other') == 'other'

def test_colored_output():
    say = Fmt()
    with say.settings(prefix='\x1b[34m', suffix='\x1b[0m'):
        assert say('this is blue!') == '\x1b[34mthis is blue!\x1b[0m'
    assert say('not blue') == 'not blue'
    blue = say.clone(prefix='\x1b[34m', suffix='\x1b[0m')
    assert blue('BLUE') == '\x1b[34mBLUE\x1b[0m'

def test_example_3():
    say = Fmt()
    items = '1 2 3'.split()

    assert say('TITLE') == 'TITLE'
    for item in items:
        assert say(item, indent=1) == '    ' + str(item)


def test_wrap():
    say = Fmt()
    assert say('abc\ndef\nghi', wrap=79) == 'abc\ndef\nghi'
    assert say("abcde abcde abcde", wrap=6) == 'abcde\nabcde\nabcde'
    assert say("abcde abcde abcde", wrap=10, indent=1) == '    abcde\n    abcde\n    abcde'

def test_vsep():
    say = Fmt(end='\n')
    assert say.title('hey', sep='-', vsep=2) == '\n\n--------------- hey ---------------\n\n\n'

def test_hr():
    say = Fmt(end='\n')
    assert say.hr(sep='-') == '----------------------------------------\n'
    assert say.hr(sep='-', vsep=2) == '\n\n----------------------------------------\n\n\n'

def test_blank_lines():
    say = Fmt(end='\n')

    assert say.blank_lines(3) == '\n' * 3

def test_numberer():
    say = Fmt(end='\n', prefix=numberer())
    assert say('this\nand\nthat') == '  1: this\n  2: and\n  3: that\n'

    say = Fmt(end='\n', prefix=numberer(template='{n:>4d}: ', start=100))

    assert say('this\nand\nthat') == ' 100: this\n 101: and\n 102: that\n'

def test_caller_fmt():

    def ucfmt(s):
        return caller_fmt(s).upper()

    x = 12
    y = 'x'

    assert ucfmt("{y!r} is {x}") == "'X' IS 12"

def test_FmtException():

    class Blooper(FmtException, ValueError):
        pass

    x = 12
    try:
        raise Blooper('{x} is cray, yo')
    except ValueError as e:
        assert e.args[0] == '12 is cray, yo'
    except Exception:
        assert False



import six
import os
import sys
from say import Text
import pytest


def test_text_basic():
    t = Text()

    x, y, s = 1, 2, 'this thing'
    t.append('{x} < {y}')
    assert t.text == '1 < 2'
    t.append('{s} x')
    t.append('and {s!r}')
    assert t.text == "1 < 2\nthis thing x\nand 'this thing'"

def test_scripting_example():

    hostname = 'server1234.example.com'
    filepath = 'ping-results.txt'

    script = Text()
    script += """
        !#/bin/bash

        # Output the results of a ping command to the given file

        ping {hostname!r} >{filepath!r}
    """

    assert str(script) == "!#/bin/bash\n\n# Output the results of a ping command to the given file\n\nping 'server1234.example.com' >'ping-results.txt'"


def test_append_raw():
    t = Text()
    x = 212
    t.append('{x}')
    t.append('{x}', interpolate=False)
    assert t.text == '212\n{x}'


def test_insert():
    x = 'Thunderdome'

    t = Text('this\nis\n{x}')
    t.insert(2, 'not')
    assert t.text == 'this\nis\nnot\nThunderdome'

    t1 = Text('this\nis\n{x}')
    t1.insert(2, 'not\nyet')
    assert t1.text == 'this\nis\nnot\nyet\nThunderdome'

    t3 = Text('this\nis\n{x}')
    t3.insert(0, 'and')
    assert t3.text == 'and\nthis\nis\nThunderdome'

    t4 = Text('this\nis\n{x}')
    t4.insert(-1, 'a')
    assert t4.text == 'this\nis\na\nThunderdome'

    t5 = Text('this\nis\n{x}')
    t5.insert(len(t5), 'is it not?')
    assert t5.text == 'this\nis\nThunderdome\nis it not?'

    y = 'is it not?'
    t6 = Text('this\nis\n{x}')
    t6.insert(len(t6), '{y}')
    assert t6.text == 'this\nis\nThunderdome\nis it not?'


def test_len():

    t = Text('1\n2\n\3\n')
    assert len(t) == 3


def test_ior():
    t = Text('one')

    x = 2
    t |= '   two {x}'
    assert t.text == 'one\n   two 2'


def test_iand():
    t = Text('one')

    x = 2
    t &= '   two {x}'
    assert t.text == 'one\n   two {x}'


def test_iadd():
    t = Text()

    t += """
        # this is a script
        # which should be left aligned and compact
            # with one indented line
    """
    assert t.text == "# this is a script\n# which should be left aligned and compact\n    # with one indented line"

    t2 = Text()
    t2 += """
        # this is a script
        # which should be left aligned and compact
            # with one indented line
    """
    assert t2.text == "# this is a script\n# which should be left aligned and compact\n    # with one indented line"


def test_adding_other_text():

    t = Text("this is good")
    u = Text("this too!")

    tu = t + u
    assert tu.text == "this is good\nthis too!"

    t += u
    assert t.text == "this is good\nthis too!"


def test_indexing():
    t = Text('this\nis\nsomething')
    assert t[0] == 'this'
    assert t[1] == 'is'
    assert t[2] == 'something'
    t[1] = 'would be'
    assert t.text == 'this\nwould be\nsomething'


def test_replace():
    t = Text('a\nb\nc\nd')
    assert t.lines == 'a b c d'.split()

    t.replace('b', 'B')
    tres = t.replace('d', 'D')
    assert t.lines == 'a B c D'.split()
    assert tres == t


def test_replace_many():
    t = Text('a\nb\nc\nd')
    assert t.lines == 'a b c d'.split()

    tres = t.replace({'b': 'BEE', 'd': 'DEE'})
    assert t.lines == 'a BEE c DEE'.split()
    assert tres == t


def test_replace_with_text():
    t = Text()
    t += 'this is\nsome\ntext here'

    middle = Text("some incredible\nawesome\nlively")
    t.replace('some', middle)
    assert t.text == 'this is\nsome incredible\nawesome\nlively\ntext here'
    assert len(t) == 5


@pytest.mark.xfail
def test_re_replace():
    t = Text('a\nb\nc\nd')
    assert t.lines == 'a b c d'.split()

    t.re_replace(r'[b]', 'B')
    t.re_replace(r'd', lambda m: m.group(0).upper())
    assert t.lines == 'a B c D'.split()

    t1 = Text('and then there were none'.split())
    t1.re_replace(r'(?P<art>the)', lambda m: '*' + m.art.upper() + '*')
    assert t1.lines == 'and *THE*n *THE*re were none'.split()


def test_set_text():
    t = Text()

    t.text = 'this\nis'
    assert t.text == 'this\nis'
    assert t.lines == ['this', 'is']

    t.lines = ['and', 'one', 'more']
    assert t.text == 'and\none\nmore'


def test_excess_newlines():
    t = Text()

    t.lines = ['too\n', 'many\n', 'newlines\n']
    assert t.text == 'too\nmany\nnewlines'
    assert t.lines == ['too', 'many', 'newlines']


def test_iter():
    t = Text('someday\nsoon')
    assert [x for x in t] == ['someday', 'soon']


def test_init():

    d = 4409

    t = Text('someday\nsoon {d}')
    assert t.text == 'someday\nsoon 4409'

    t = Text(['someday', 'soon {d}'])
    assert t.text == 'someday\nsoon 4409'

    b = 'boy'

    t2 = Text('hey {b}')
    assert t2.text == 'hey boy'

    t3 = Text('hey {b}', interpolate=False)
    assert t3.text == 'hey {b}'

    t4 = Text("""
        Now this
        is more
        like it!
    """)
    assert t4.text == "Now this\nis more\nlike it!"

    t5 = Text(dedent=False, data="""
        Now this
        is more
        like it!
    """)
    assert t5.text == "\n        Now this\n        is more\n        like it!\n    "


def test_str():
    t = Text('someday\nsoon')

    assert str(t) == 'someday\nsoon'


def test_repr():
    t = Text('someday\nsoon')

    import re
    assert re.match('Text\(\d+, \d+ lines\)', repr(t))



"""Interpolating string formatter. """

import string
import inspect
import sys
import os
import io
import re
from stuf import chainstuf
from options import Options, OptionsContext, Transient, Prohibited
import six
import textwrap
from say.util import *
from say.styling import *
from say.vertical import Vertical, vertical

if six.PY3:
    unicode = str

### Workhorse functions

sformatter = string.Formatter()  # piggyback Python's format() template parser

def populate_style(style_name, styles):
    if style_name.startswith(("'", '"')):
        style_name = style_name[1:-1]
    if style_name not in styles:
        styles[style_name] = autostyle(style_name)


def _sprintf(arg, caller, styles=None, override=None):
    """
    Format the template string (arg) with the values seen in the context of the
    caller. If override is defined, it is a Mapping providing additional values
    atop those of the local context.
    """

    def seval(s):
        """
        Evaluate the string s in the caller's context. Return its value.
        """
        try:
            localvars = caller.f_locals if override is None \
                                        else chainstuf(override, caller.f_locals)
            return eval(s, caller.f_globals, localvars)
        except SyntaxError:
            raise SyntaxError("syntax error when formatting '{0}'".format(s))

    def parse_style(fs, styles):
        """
        Get the style component out of the format string, if any.
        """
        if "style" in fs:
            fs_parts = fs.split(',')
            raw_fs_parts = []
            for fsp in fs_parts:
                if fsp.startswith('style='):
                    style_name = fsp[6:].strip(QUOTE_DELIM_STR)
                else:
                    raw_fs_parts.append(fsp)
            populate_style(style_name, styles)
            return style_name, raw_fs_parts
        else:
            return None, fs

        # TODO: Replace this parser with something more professional
        # TODO: join=

    if is_string(arg):
        arg = unicode(arg) if six.PY2 and isinstance(arg, str) else arg
        parts = []
        for (literal_text, field_name, format_spec, conversion) in sformatter.parse(arg):
            parts.append(literal_text)
            if field_name is not None:

                style_name, raw_format_spec = parse_style(format_spec, styles)
                format_str = six.u("{0") + \
                                   ("!" + conversion if conversion else "") + \
                                   (":" + raw_format_spec if raw_format_spec else "") + "}"
                field_value = seval(field_name)
                formatted = format_str.format(field_value)
                if style_name and style_name in styles:
                    formatted = styles[style_name](formatted)

                parts.append(formatted)
        return ''.join(parts)
    else:
        return str(seval(str(arg)))


### Core Say class

class Say(object):

    """
    Say encapsulates printing functions. Instances are configurable, and callable.
    """

    options = Options(
        indent=0,           # indent level (if set to None, indentation is turned off)
        indent_str='    ',  # indent string for each level
        prefix='',          # prefix each line with this (string or string generator)
        suffix='',          # suffix each line with this (string or string generator)
        files=[sys.stdout], # where is output headed? a list of write() able objects
        wrap=None,          # column to wrap lines to, if any
        sep=' ',            # separate args with this (Python print function compatible)
        vsep=None,          # vertical separation
        end='\n',           # end output with this (Python print function compatible)
        silent=False,       # be quiet
        style=None,         # name of style in which to wrap entire output line
        styles={},          # style dict
        _callframe=Transient, # frome from which the caller was calling
    )

    options.magic(
        indent = lambda v, cur: cur.indent + int(v) if isinstance(v, str) else v,
        files  = lambda v, cur: opened(v)
    )

    def __init__(self, **kwargs):
        """
        Make a say object with the given options.
        """
        self.options = Say.options.push(kwargs)

    @staticmethod
    def escape(s):
        """
        Double { and } characters in a string to 'escape' them so ``str.format``
        doesn't treat them as template characters. NB This is NOT idempotent!
        Escaping more than once (when { or } are present ) = ERROR.
        """
        return s.replace('{', '{{').replace('}', '}}')

    def hr(self, sep=six.u('\u2500'), width=40, **kwargs):
        """
        Print a horizontal line. Like the HTML hr tag. Optionally
        specify the width, character repeated to make the line, and vertical separation.

        Good options for the separator may be '-', '=', or parts of the Unicode
        box drawing character set. http://en.wikipedia.org/wiki/Box-drawing_character
        """
        opts = self.options.push(kwargs)
        # no interpolation required, so no caller frame introspection
        line = sep * width
        return self._output(line, opts)

    def title(self, name, sep=six.u('\u2500'), width=15, **kwargs):
        """
        Print a horizontal line with an embedded title.
        """
        opts = self.options.push(kwargs)
        opts.setdefault('_callframe', inspect.currentframe().f_back)

        formatted = _sprintf(name, opts._callframe, opts.styles) if is_string(name) else str(name)
        bars = sep * width
        line = ' '.join([bars, formatted, bars])
        return self._output(line, opts)

    def blank_lines(self, n, **kwargs):
        """
        Output N blank lines ("vertical separation"). Unlike other methods, this
        does not obey normal vertical separation rules, because it is about
        explicit vertical separation. If it obeyed vsep, it would usually gild
        the lily (double space).
        """
        if not n:
            return
        opts = self.options.push(kwargs)
        # no interpolation required, so no caller frame introspection
        opts.vsep = None
        return self._output([''] * n, opts)

    def set(self, **kwargs):
        """
        Permanently change the reciver's settings to those defined in the kwargs.
        An update-like function.
        """
        self.options.set(**kwargs)

    def setfiles(self, files):
        """
        Set the list of output files. ``files`` is a list. For each item, if
        it's a real file like ``sys.stdout``, use it. If it's a string, assume
        it's a filename and open it for writing.
        """
        def opened(f):
            """
            If f is a string, consider it a file name and return it, ready for writing.
            Else, assume it's an open file. Just return it.
            """
            return io.open(f, "w") if is_string(f) else f
            # NB use io.open not plain open to get auto-encoding to UTF-8 in Python 2

        self.options.files = [opened(f) for f in files]
        return self

        # TBD: Turn this into 'magical' attribute set

    def settings(self, **kwargs):
        """
        Open a context manager for a `with` statement. Temporarily change settings
        for the duration of the with.
        """
        return SayContext(self, kwargs)

    def clone(self, **kwargs):
        """
        Create a new Say instance whose options are chained to this instance's
        options (and thence to Say.options). kwargs become the cloned instance's
        overlay options.
        """
        cloned = self.__class__()
        cloned.options = self.options.push(kwargs)
        return cloned

    def style(self, *args, **kwargs):
        """
        Define a style.
        """
        for k,v in kwargs.items():
            if isinstance(v, six.string_types):
                self.options.styles[k] = autostyle(v)
            else:
                self.options.styles[k] = v


    def __call__(self, *args, **kwargs):
        """
        Primary interface. say(something)
        """
        opts = self.options.push(kwargs)
        opts.setdefault('_callframe', inspect.currentframe().f_back)

        formatted = [ _sprintf(arg, opts._callframe, opts.styles) if is_string(arg) else str(arg)
                      for arg in args ]
        return self._output(opts.sep.join(formatted), opts)

    def __gt__(self, arg):
        """
        Simple, non-functional call. Experimental.
        """
        opts = self.options.push({})
        opts.setdefault('_callframe', inspect.currentframe().f_back)

        formatted = [ _sprintf(arg, opts._callframe, opts.styles) if is_string(arg) else str(arg) ]
        return self._output(opts.sep.join(formatted), opts)

    def _return_value(self, outstr, opts):
        """
        Prepare a quality return value. Manages encoding of string and
        stripping of final newline, if desired, based on opts.return_encoded
        and opts.return_strip_newline.
        """
        ret_encoding = opts.encoding if opts.return_encoded is True else opts.encoded
        ret_encoded = encoded(outstr, ret_encoding)
        if opts.return_strip_newline:
            ret_encoded = ret_encoded[:-1] if ret_encoded.endswith('\n') else ret_encoded
        return ret_encoded


    def _outstr(self, data, opts):
        """
        Given result text, format it. ``data`` may be either a
        list of lines, or a composed string. NB: Don't feed it a list of strings,
        some of which contain newlines; that will break its assumptions.
        """
        datalines = data if isinstance(data, list) else data.splitlines()

        if opts.indent or opts.wrap or opts.prefix or opts.suffix or opts.vsep or opts.style:
            indent_str = opts.indent_str * opts.indent
            if opts.wrap:
                datastr = '\n'.join(datalines)
                wrap_width = opts.wrap - len(indent_str) - len(opts.prefix) - len(opts.suffix)
                wrappedlines = textwrap.wrap(datastr,
                                          width=wrap_width,
                                          replace_whitespace=False,
                                          initial_indent='',
                                          subsequent_indent='')
                datalines = []
                for line in wrappedlines:
                    datalines.extend(line.splitlines())
            if opts.style:
                styler = opts.styles.get(opts.style, None)
                if not styler:
                    styler = opts.styles.setdefault(opts.style, autostyle(opts.style))
                datalines = [styler(line) for line in datalines]
            if opts.indent:
                datalines = [ indent_str + line for line in datalines ]
            vbefore, vafter = vertical(opts.vsep).render()
            datalines = vbefore + datalines + vafter
            if opts.prefix or opts.suffix:
                datalines = [''.join([next_str(opts.prefix), line, next_str(opts.suffix)])
                              for line in datalines]
            outstr = '\n'.join(datalines)
        else:
            outstr = '\n'.join(data) if isinstance(data, list) else data
        # by end of indenting, dealing with string only

        # prepare and emit output
        if opts.end is not None:
            outstr += opts.end
        return outstr


    def _output(self, data, opts):
        """
        Construct the output string and write it to files.
        """
        if opts.silent:
            return
        else:
            outstr = self._outstr(data, opts)
            for f in opts.files:
                f.write(outstr)


class SayContext(OptionsContext):
    """
    Context helper to support Python's with statement.  Generally called
    from ``with say.settings(...):``
    """
    pass


class Fmt(Say):
    """
    A type of Say that returns its result, rather than writes it
    to files.
    """

    options = Say.options.copy().add(
        files=Prohibited,   # no files needed
        end=None,           # no extra newline
        silent=Prohibited,  # Fmt is never silent
    )

    def __init__(self, **kwargs):
        self.options = Fmt.options.push(kwargs)
        self.options.styles = say.options.styles  # styles are idiosyncratially shared


    def _output(self, data, opts):
        """
        Construct the output string and return it.
        """
        return self._outstr(data, opts)


### Define default callables

say = Say()
fmt = Fmt()


### Helpers

def caller_fmt(*args, **kwargs):
    """
    Like ``fmt()``, but iterpolating strings not from the caller's context, but
    the caller's caller's context. It sounds uber meta, but it helps easily make
    other routines be able to do what ``fmt()`` can do.
    """
    kwargs.setdefault('_callframe', inspect.currentframe().f_back.f_back)
    return fmt(*args, **kwargs)


class FmtException(Exception):
    """
    An exception class that formats its arguments in the calling context.
    Also, an example of how ``caller_fmt`` can be used.
    """
    def __init__(self, message):
        formatted = caller_fmt(message)
        super(FmtException, self).__init__(formatted)


class numberer(object):
    """
    A class-based factory for numbering generators. Rather like Python 2's
    ``xrange`` (or Python 3's ``range``), but intended to number lines in a file
    so it uses natural numbers starting at 1, not the typical Python 'start at
    0'. Also, returns formatted strings, not integers. Improves on what's
    possible as a functional generator on numberer because it can compute and
    return its own length.
    """

    def __init__(self, start=1, template="{n:>3}: ", length=None):
        """
        Make a numberer.
        """
        self.n = start
        self.template = template
        self.length = None
        self._formatted = None

    def __next__(self):
        """
        Return the next numbered template.
        """
        t = self.template
        result = t.format(n=self.n) if self._formatted is None else self._formatted
        self._formatted = None
        self.n += 1
        return result

    next = __next__   # for Python 2

    def __len__(self):
        """
        What is the string length of the instantiated template now? NB This can change
        over time, as n does. Fixed-width format strings limit how often it can change
        (to when n crosses a power-of-10 boundary > the fixed template length
        can accomodate). This implementation saves the templated string its has created
        for reuse. If self.length is set, however, this computation is not done, and
        self.length is assumed to be the right answer.
        """
        if self.length is not None:
            return self.length
        else:
            t = self.template
            result = t.format(n=self.n)
            self._formatted = result
            return len(result)

    # TODO: provide an ANSI-control-character independent version of string length that
    # can provide a "logical string length"



"""
Module to handle styled printing
"""

import six
from colors import color as _color, COLORS, STYLES
from stuf import stuf
if six.PY3:
    unicode = str

QUOTE_DELIM = ('"', "'")
QUOTE_DELIM_STR = ''.join(QUOTE_DELIM)

# TODO: implement a persistent Style object that say.style will return
# TODO: implement own six-like features, rather than hanging between six and custom code

class optdict(dict):
    """
    dict subclass that only initializes those keys where there's a non-empty value.
    Convenient for creating kwarg dicts that don't have to define default values.
    """
    def __init__(self, **kwargs):
        dict.__init__(self)
        for k,v in kwargs.items():
            if v:
                self[k] = v


class StyleDef(object):
    """
    Class defining a formatting style.
    """

    def __init__(self, *args, **kwargs):
        """
        Create a new StyleDef, remembering its settings.
        """
        self.name = kwargs.get('name', None)
        self.sdef = styledef(*args, **kwargs)
        self.join = kwargs.get('join', None)

    def __call__(self, item):
        """
        Apply this style to the given item.
        """
        result = self.join(item) if self.join else item
        result = _color(result, **self.sdef) if self.sdef else result
        return result

def styledef(*args, **kwargs):
    """
    Parses a style definition as given in args and kwargs.
    Return a dict that defines a given style definition. If it's an ANSI style
    being defined, whatever color is named first (if any) is assumed to be the
    foreground, the second the background. ANSI display styles (like 'bold') can
    be named anywhere.
    """
    kw = stuf()
    for arg in args:
        # print "ARG:", arg
        arg = arg.replace('+', '|').replace(',', '|').lower()
        parts = [p.strip() for p in arg.split('|')]
        fg, bg, styles = None, None, []
        for p in parts:
            if p in COLORS:
                if not fg:
                    fg = p
                elif not bg:
                    bg = p
                else:
                    raise ValueError('only fg and bg colors!')
            elif p in STYLES:
                styles.append(p)
        kw.update(optdict(fg=fg, bg=bg, style='|'.join(styles) if styles else None))
    kw.update(kwargs)
    # print "KW:", kw
    return kw

def autostyle(*args, **kwargs):
    """
    Return a lambda that will later format a string with the given styledef.
    """
    sdef = styledef(*args, **kwargs)
    return lambda x: color(x, **sdef)

def color(item, **kwargs):
    """
    Like colors.color, except auto-casts to Unicode string if need be.
    """
    item_u = unicode(item) if not isinstance(item, six.string_types) else item
    return _color(item_u, **kwargs)

def styled(item, *args, **kwargs):
    """
    Like color, but can also include style names.
    Kind of an immediate-mode version of autostyle.
    """
    sdef = styledef(*args, **kwargs)
    return color(item, **sdef)



"""
Module to assist in the super-common operation of
joining values in sequences into strings.
"""

import re
from options import Options
import six

def is_string(v):
    return isinstance(v, six.string_types)

def stringify(v):
    """
    Return a string. If it's already a string, just return that.
    Otherwise, stringify it.
    """
    return v if is_string(v) else str(v)

def blanknone(v):
    """
    Return a value, or empty string if it's None.
    """
    return '' if v is None else v

joiner_options = Options(
    sep = ', ',    # separator between items
    twosep=None,   # separator between items if only two
    lastsep=None,  # separator between penultimate and final item
    quoter=None,   # quoter for individual items
    endcaps=None,  # quoter for entire joined sequence
    prefix=None,   # prefix for entire joined, endcaped sequence
    suffix=None    # suffix for entire joined, endcaped sequence
)

def joiner(seq, **kwargs):
    """
    Join the items of a sequence into a string. Implicitly stringifies any
    not-string values. Allows specification of the separator between items (and
    a special case for the last separator). Allows each item to be optionally
    quoted by a function, and the entire list to be optionally quoted with an
    endcaps function. A separate suffix and prefix may also be provdied.
    """

    opts = joiner_options.push(kwargs)

    def prep(v):
        """
        Prepare an item by stringifying and optionally quoting it.
        """
        s = stringify(v)
        return opts.quoter(s) if opts.quoter else s

    seqlist = list(seq)
    length = len(seqlist)
    if length == 0:
        core = ''
    elif length == 1:
        core = prep(seqlist[0])
    elif length == 2 and opts.twosep:
        sep = opts.twosep if opts.twosep is not None else opts.sep
        core = sep.join(prep(v) for v in seqlist)
    else:
        start = [ prep(v) for v in seqlist[:-1] ]
        final = prep(seqlist[-1])
        if opts.lastsep is None:
            opts.lastsep = opts.sep
        core = opts.lastsep.join([ opts.sep.join(start), final])
    capped = opts.endcaps(core) if opts.endcaps else core
    if opts.prefix or opts.suffix:
        affixed = ''.join([ blanknone(opts.prefix), capped, blanknone(opts.suffix) ])
    else:
        affixed = capped
    return affixed

def word_join(seq, **kwargs):
    """
    Slightly specific joiner for words. Returns by default an Oxford comma list,
    but accepts all the same options as joiner.
    """
    kwargs.setdefault('sep', ', ')
    kwargs.setdefault('twosep', ' and ')
    kwargs.setdefault('lastsep', ', and ')
    return joiner(seq, **kwargs)

join = joiner
and_join = word_join

def or_join(seq, **kwargs):
    """
    A, B, or C.
    """
    kwargs.setdefault('sep', ', ')
    kwargs.setdefault('twosep', ' or ')
    kwargs.setdefault('lastsep', ', or ')
    return joiner(seq, **kwargs)

#def is_sequence(arg):
#    return (not hasattr(arg, "strip") and
#            hasattr(arg, "__getitem__") or
#            hasattr(arg, "__iter__"))



concat_options = joiner_options.add(
    sep = '', # combine with no separator, by default
)

def concat(*args, **kwargs):
    six.print_("kwargs:", kwargs)
    six.print_("concat_options:", concat_options)
    opts = concat_options.push(kwargs)
    six.print_("opts:", opts)
    six.print_("args:", args)

    args=list(args)
    six.print_("listed args:", args)

    if len(args) == 1 and is_sequence(args):
        args = list(args[0])
    six.print_("args:", args)
    six.print_("opts:", opts)
    return joiner(args, **opts)  # problem here in Py3x

items_options = Options(
    sep = "\n",  # separator between items
    fmt = "{key}: {value!r}",
    header = None,   # header for entire list
    footer = None    # footer for entire list
)

def iter_items(items):
    if hasattr(items, 'items'):  # dict or mapping
        for k, v in items.items():
            yield k, v
    else:
        for k, v in enumerate(items):
            yield k, v

def items(seq, **kwargs):
    opts = items_options.push(kwargs)

    formatted_items = [ opts.fmt.format(key=k, value=v) for k,v in iter_items(seq) ]
    items_str = opts.sep.join(formatted_items)
    if opts.header or opts.footer:
        parts = []
        if opts.header:
            parts.extend([opts.header, opts.sep])
        parts.append(items_str)
        if opts.footer:
            parts.extend([opts.sep, opts.footer])
        items_str = ''.join(parts)
    return items_str




from say.joiner import *
import pytest, sys

### Helper functions

def single(s):
    return "'" + s + "'"

def brackets(s):
    return "[" + s + "]"


### Tests

def test_oxford():
    assert word_join([1,2,3]) == '1, 2, and 3'

def test_heathen():
    assert word_join([1,2,3], lastsep=" and ") == '1, 2 and 3'

def test_quoted():
    assert word_join([], quoter=single) == ""
    assert word_join([1], quoter=single) == "'1'"
    assert word_join([1,2], quoter=single) == "'1' and '2'"
    assert word_join([1,2,3], quoter=single) == "'1', '2', and '3'"

def test_listy():
    assert joiner([], quoter=single, endcaps=brackets) == "[]"
    assert joiner([1], quoter=single, endcaps=brackets) == "['1']"
    assert joiner([1,2], quoter=single, endcaps=brackets) == "['1', '2']"
    assert joiner([1,2,3], quoter=single, endcaps=brackets) == "['1', '2', '3']"

def test_sep():
    assert joiner([], sep='|') == ''
    assert joiner([1], sep='|') == '1'
    assert joiner([1,2], sep='|') == '1|2'
    assert joiner([1,2,3], sep='|') == '1|2|3'

def test_twosep_and_lastsep():
    assert joiner([1,2,3,4], sep='|', lastsep='+') == '1|2|3+4'

    assert joiner([], sep='|', twosep='*', lastsep='+') == ''
    assert joiner([1], sep='|', twosep='*', lastsep='+') == '1'
    assert joiner([1,2], sep='|', twosep='*', lastsep='+') == '1*2'
    assert joiner([1,2,3], sep='|', twosep='*', lastsep='+') == '1|2+3'
    assert joiner([1,2,3,4], sep='|', twosep='*', lastsep='+') == '1|2|3+4'

def test_no_twostep():
    assert joiner([], sep='|', twosep=None, lastsep='+') == ''
    assert joiner([1], sep='|', twosep=None, lastsep='+') == '1'
    assert joiner([1,2], sep='|', twosep=None, lastsep='+') == '1+2'
    assert joiner([1,2,3], sep='|', twosep=None, lastsep='+') == '1|2+3'
    assert joiner([1,2,3,4], sep='|', twosep=None, lastsep='+') == '1|2|3+4'


def test_concat():
    assert concat(4,5,6) == '456'
    # assert concat(range(3)) == '012'
    # assert concat('a','b','c') == 'abc'

def test_and_join():
    assert and_join([]) == ''
    assert and_join([1]) == '1'
    assert and_join([1,2]) == '1 and 2'
    assert and_join([1,2,3]) == '1, 2, and 3'
    assert and_join([1,2,3,4]) == '1, 2, 3, and 4'

def test_or_join():
    assert or_join([]) == ''
    assert or_join([1]) == '1'
    assert or_join([1,2]) == '1 or 2'
    assert or_join([1,2,3]) == '1, 2, or 3'
    assert or_join([1,2,3,4]) == '1, 2, 3, or 4'

def test_items():
    assert items([1,2,3,'string']) == "0: 1\n1: 2\n2: 3\n3: 'string'"

    assert items([1,2,3,'string'], header="---") == "---\n0: 1\n1: 2\n2: 3\n3: 'string'"
    assert items([1,2,3,'string'], footer="===") == "0: 1\n1: 2\n2: 3\n3: 'string'\n==="
    assert items([1,2,3,'string'], header='---', footer="===") == "---\n0: 1\n1: 2\n2: 3\n3: 'string'\n==="


    try:
        from collections import OrderedDict
        answer = "this: something\nthat: else\nplus: additionally"
    except ImportError:
        OrderedDict = dict # py26
        answer = "this: something\nplus: additionally\nthat: else"
        # output in different order because unordred dict
    od = OrderedDict()
    od['this'] = 'something'
    od['that'] = 'else'
    od['plus'] = 'additionally'

    assert items(od, fmt="{key}: {value}") == answer

    assert items(od, fmt="{key}: {value}", header="KEY: VALUE") == \
        "KEY: VALUE\n" + answer



"""
A range of say.text.Text lines that can be edited on their own,
effecting the underlying lines. If the underlying text object
changes (especially changing the number of lines somewhere
before or overlapping whether the text range lies), TextRange
will probably become confused. However, edits made through the
TextRange are completely fine.
"""

# TextRange shows the difficulty of 'sub object' references in
# Python. On the one hand, if t is a Text object, t[1:4] should
# return a list of strings for those lines. OTOH, maybe it should
# return an abstracted TextRange object. But then if the underlying
# Text object changes, how to have the TextRange then adjust? I.e.
# if a few lines are prepended to t, shouldn't the text range still
# refer to the new lines, albeit at a higher index range? Building
# this invovles dependencies, and Text knowing about TextRange, and
# vice versa. For the meanwhile, lacking a generalized dependency
# mechanism (not to mention a generalized lvalue mechanism),
# just going to make TextRange be static.

# Also note how TextRange is not really a subclass of Text, though
# logically it could, or should, or sorta wants to be. But even if
# it were, we'd still have to do lots of the copy-paste-modify
# style of programming here.

# For further investigation:
# http://pypi.python.org/pypi/Cellulose

from say import Text
import inspect
import sys
import re
from simplere import Re

_PY3 = sys.version_info[0] > 2
if _PY3:
    basestring = str


class TextRange(object):

    def __init__(self, text=None, start=None, stop=None):
        """
        Create a TextRange object on the given text, over the given range.
        Just as with Python list specifications, the stop value should be
        one more than the last index.
        """
        self._text = text
        if stop is None:
            stop = len(text)
        elif stop < 0:
            stop = len(text) + stop
        if start is None:
            start = 0
        elif start < 0:
            start = len(text) + start
        self._indices = slice(start, stop)

    def __iadd__(self, data):
        """
        In-place add the text or lines contained in data, with auto-dedent.
        """
        caller = inspect.currentframe().f_back
        self._insert(self._indices.stop, data,
                     caller, interpolate=True, dedent=True)
        return self

    def __ior__(self, data):
        """
        In-place add the text or lines contained in data, with NO auto-dedent.
        """
        caller = inspect.currentframe().f_back
        self._insert(self._indices.stop, data,
                     caller, interpolate=True, dedent=False)
        return self

    def __iand__(self, data):
        """
        In-place add the text or lines contained in data, with NO auto-dedent
        and NO iterpolation.
        """
        self._insert(self._indices.stop, data,
                     None, interpolate=False, dedent=False)
        return self

    def append(self, data, caller=None, interpolate=True):
        caller = caller or inspect.currentframe().f_back if interpolate else None
        self._insert(self._indices.stop, data, caller=caller, interpolate=interpolate)

    def extend(self, data, caller=None, interpolate=True, dedent=True):
        caller = caller or inspect.currentframe().f_back if interpolate else None
        self._insert(self._indices.stop, data, caller, interpolate, dedent)

    def _base_index(self, n):
        """
        Check to the given index n to ensure it's within the range of lines.
        """
        index = self._indices.start + n if n >= 0 else self._indices.stop + n
        if self._indices.start <= index < self._indices.stop:
            return index
        raise IndexError('index {0} ({1} in underlying Text) out of range'.format(n, index))

    def _insert(self, n, data, caller=None, interpolate=True, dedent=True):
        """
        Insert into the underlying Text at a point relative to the underlying text indices.
        """
        newlines = self._data_to_lines(data)
        caller = caller or inspect.currentframe().f_back if interpolate else None
        self._text.insert(n, newlines, caller, interpolate, dedent)
        self._adjust_indices(newlines, replacing=False)

    def insert(self, n, data, caller=None, interpolate=True, dedent=True):
        """
        Insert into the underlying Text at a point relative to the TextRange's
        indices.
        """
        caller = caller or inspect.currentframe().f_back if interpolate else None
        self._insert(self._indices.start + n, data, caller, interpolate, dedent)

    def __getitem__(self, n):
        index = self._base_index(n)
        return self._text._lines[index]

    def __setitem__(self, n, value):
        index = self._base_index(n)
        self._text._lines[index] = value.rstrip('\n')

    def __len__(self):
        return self._indices.stop - self._indices.start

    def __iter__(self):
        return iter(self._text._lines[self._indices])

    def _data_to_lines(self, data):
        if isinstance(data, list):
            return [line.rstrip('\n') for line in data]
        else:
            return data.splitlines()

    @property
    def text(self):
        return '\n'.join(self._text._lines[self._indices])

    @text.setter
    def text(self, data):
        newlines = self._data_to_lines(data)
        self._text._lines[self._indices] = newlines
        self._adjust_indices(newlines)

    def _adjust_indices(self, newlines, replacing=True):
        """
        If a TextRange is modified, we adjust ._indices to reflect the length
        of the new text. If replacing, assume the new lines replace the given
        text range; if not replacing, adding to the text range.
        """
        newlen = len(newlines)
        newstop = self._indices.start + newlen if replacing else self._indices.stop + newlen
        if self._indices.stop != newstop:
            self._indices = slice(self._indices.start, newstop)

    @property
    def lines(self):
        return self._text._lines[self._indices]

    @lines.setter
    def lines(self, newlines):
        newlines = [line.rstrip('\n') for line in newlines]
        self._text._lines[self._indices] = newlines
        self._adjust_indices(newlines)

    def __str__(self):
        return self.text

    def __repr__(self):
        return 'TextRange({0}, {1}:{2} of {3})'.format(id(self),
                self._indices.start, self._indices.stop, id(self._text))

    def replace(self, target, replacement=None):
        """
        Replace all instances of the target string with the replacement string.
        Works in situ, contra str.replace().
        """
        if target and isinstance(target, dict) and replacement is None:
            for (targ, repl) in target.items():
                self.replace(targ, repl)
        else:
            for i, line in enumerate(self._text._lines[self._indices], start=self._indices.start):
                self._text._lines[i] = line.replace(target, replacement)

        # TODO: should lines be recalibrated, if there are any \n in
        # replacement?

    def re_replace(self, target, replacement):
        """
        Regular expression replacement. Target is either compiled re object or
        string that will be compiled into one. Replacement is either string or
        function that takes re match object as a parameter and returns replacement
        string. Intead of Python's stock ``_sre.SRE_Match`` objects, returns a
        ``ReMatch``, which adds the ability to directly access via indices
        """
        target = Re(target) if not isinstance(target, Re) else target
        for i, line in enumerate(self._text._lines[self._indices], start=self._indices.start):
            self._text._lines[i] = target.sub(replacement, line)

    def copy(self):
        """
        Make a copy.
        """
        raise NotImplementedError('not sure what makes sense')

    def read_from(self, filepath, interpolate=True, dedent=True):
        """
        Reads lines from the designated file, appending them to the end of the
        given TextRange. By default, interpolates and dedents any {}
        expressions.
        """
        lines = open(filepath).read().splitlines()
        caller = inspect.currentframe().f_back
        self.extend(lines, caller, interpolate=interpolate, dedent=dedent)
        return self

    def write_to(self, filepath, append=False, encoding='utf-8'):
        mode = "a" if append else "w"
        with open(filepath, "w") as f:
            f.write(self.text.encode(encoding))


"""
Define veritcal separation.
"""

from mementos import MementoMetaclass, with_metaclass
import sys
_PY3 = sys.version_info[0] == 3
if _PY3:
    unicode = str

class Vertical(with_metaclass(MementoMetaclass, object)):
    """
    Specification for vertical spacing. Object creation is
    memoized for efficiency.
    """
    def __init__(self, before=0, after=0, bstr='', astr=''):
        """
        Make one.
        """
        self.before = before
        self.after = after
        self.bstr = bstr
        self.astr = astr

    def render(self):
        """
        Return a list-of-lines representation of the receiver.
        """
        before = [self.bstr] * self.before
        after = [self.astr] * self.after
        return (before, after)

    # TODO: memoize ``render`` output. Or does that crush the mementos caching strategy?
    # if changed, would have to invalidate the mementos cache?

    def __repr__(self):
        """
        Pretty representation, yo.
        """
        keys = ['before', 'after', 'bstr', 'astr']
        guts = ', '.join(['{0}={1!r}'.format(k, self.__dict__[k]) for k in keys])
        return "{0}({1})".format(self.__class__.__name__, guts)

def vertical(first=0, second=None):
    """
    Permissive factory function for vertical separation.
    """
    if not first and second is None:
        return Vertical(0,0,'','')
    elif isinstance(first, tuple) and second is None:
        return Vertical(*first)
    if isinstance(first, Vertical):
        return first
    before = first
    after  = second if second is not None else first
    return Vertical(before, after)




from say.vertical import *
import pytest

def test_basic():
    assert Vertical(0,0).render() == ([], [])
    assert Vertical(1,1).render() == ([''], [''])
    assert Vertical(1,2).render() == ([''], ['', ''])
    assert Vertical(2,1).render() == (['', ''], [''])
    assert Vertical(2,3).render() == (['', ''], ['', '', ''])

def test_strings():
    assert Vertical(0,0,'x','y').render() == ([], [])
    assert Vertical(1,1,'x','y').render() == (['x'], ['y'])
    assert Vertical(1,2,'x','y').render() == (['x'], ['y', 'y'])
    assert Vertical(2,1,'x','y').render() == (['x', 'x'], ['y'])
    assert Vertical(2,3,'x','y').render() == (['x', 'x'], ['y', 'y', 'y'])
    assert Vertical(2,0,'x','y').render() == (['x', 'x'], [])
    assert Vertical(0,3,'x','y').render() == ([], ['y', 'y', 'y'])

def test_memoization():
    assert Vertical(0,0,'x','y') is Vertical(0,0,'x','y')
    assert Vertical(0,3,'x','y') is Vertical(0,3,'x','y')
    assert Vertical(before=1, after=1, bstr='', astr='') is Vertical(before=1, after=1, bstr='', astr='')

@pytest.mark.skipif('True')
def test_vertical():
    assert vertical() is vertical(0)
    assert vertical(0) is Vertical(0,0,'','')
    assert vertical(1) is Vertical(1,1,'','')
    assert vertical(1,2) is Vertical(1,2,'','')



"""
Test separable utility functions used in say
"""

import six
from say.util import *
import pytest


def test_is_string():

    assert is_string("")
    assert is_string("This")
    assert is_string(six.u("this"))
    assert not is_string(1)
    assert not is_string(None)
    assert not is_string([1, 2, 3])
    assert not is_string(['a', 'b', 'c'])

@pytest.mark.xfail
def test_opened():
    raise NotImplementedError('TBD')

def test_encoded():

    tests = {
        ('this', 'utf-8'): six.b('this'),
        (six.u('this'), 'utf-8'): six.b('this'),
        (six.u('this\u2012and'), 'utf-8'): six.b('this\xe2\x80\x92and'),
        (six.u('this-and'), 'base64'): 'dGhpcy1hbmQ=\n',
    }

    for data, answer in tests.items():
        (text, encoding) = data
        assert encoded(text, encoding) == answer


def test_flatten(*args):

    tests = [
        (1,       [1]),
        ([1],     [1]),
        ('one',   ['one']),
        (['one'], ['one']),
        ([2, 3, 4], [2, 3, 4])

    ]

    for data, answer in tests:
        assert [x for x in flatten(data)] == answer

def test_next_str():

    def gen():
        n = 1
        while True:
            yield str(n)
            n += 1

    g = gen()

    for i in range(1,5):
        assert next_str(g) == str(i)

    assert next_str(None) == ''

    for gg in [ 1, 1.1, 'string', list, [1,2,3], {'a':'A'} ]:
        assert next_str(gg) == str(gg)

@pytest.mark.skipif('True')
def test_csv_split():
    # no quotes
    assert csv_split('x,y,z') == ['x', 'y', 'z']

    # double quotes
    assert csv_split('x,y="simple",z') == ['x', 'y="simple"', 'z']
    assert csv_split('x,y="internal comma,",z') == ['x', 'y="internal comma,"', 'z']

    # single quotes
    assert csv_split("x,y='simple',z") == ['x', "y='simple'", 'z']
    assert csv_split("x,y='internal comma,',z") == ['x', "y='internal comma,'", 'z']



from say.core import say, Say, fmt, Fmt, caller_fmt, FmtException, numberer
from say.styling import color, COLORS, STYLES, autostyle, styled
from say.vertical import Vertical
from say.text import Text
from say.version import __version__


import six
import os
from say import Text
from say.textrange import TextRange


def test_basic():
    t = Text('a\nb\nc\nd')
    assert t.lines == 'a b c d'.split()

    tr = TextRange(t, 1, 3)
    assert tr.text == 'b\nc'
    assert tr.lines == ['b', 'c']

    tr += 'hey!'
    assert t.lines == 'a b c hey! d'.split()

    assert tr.text == 'b\nc\nhey!'
    assert tr.lines == ['b', 'c', 'hey!']


def test_indexing():

    t = Text('a\nb\nc\nd')
    assert t.lines == 'a b c d'.split()

    tr = TextRange(t, 1, 3)
    assert tr.text == 'b\nc'
    assert tr.lines == ['b', 'c']

    assert tr[0] == 'b'
    assert tr[1] == 'c'

    from pytest import raises

    with raises(IndexError):
        assert tr[2] == 'd'


def test_replace():
    t = Text('a\na\na\nb\na')
    assert t.lines == 'a a a b a'.split()

    tr = TextRange(t, 2, 4)
    assert tr.lines == 'a b'.split()

    tr.replace('a', 'A')
    assert tr.lines == 'A b'.split()
    assert t.lines == 'a a A b a'.split()


def test_replace_many():
    t = Text('a\na\na\nb\na')
    assert t.lines == 'a a a b a'.split()

    tr = TextRange(t, 2, 4)
    assert tr.lines == 'a b'.split()

    tr.replace({'a': 'A', 'b': 'BEE'})
    assert tr.lines == 'A BEE'.split()
    assert t.lines == 'a a A BEE a'.split()


def test_re_replace():
    t = Text('a\na\na\nb\na')
    assert t.lines == 'a a a b a'.split()

    tr = TextRange(t, 2, 4)
    assert tr.lines == 'a b'.split()

    tr.re_replace(r'[b]', 'B')
    tr.re_replace(r'a', lambda m: m.group(0).upper())
    assert t.lines == 'a a A B a'.split()


def test_interpolation():
    x = 21
    t = Text("Joe,\nthis is {x}")
    assert t.text == 'Joe,\nthis is 21'

    tr = TextRange(t, 1, 3)

    tr += 'and {x}'
    assert tr.text == 'this is 21\nand 21'

    tr &= 'and {x}'
    assert tr.text == 'this is 21\nand 21\nand {x}'

    assert t.text == 'Joe,\nthis is 21\nand 21\nand {x}'


def test_replace():
    t = Text('a\nb\nc\nd')
    assert t.lines == 'a b c d'.split()
    tr = TextRange(t, 1, 3)
    assert tr.lines == ['b', 'c']

    tr.replace('b', 'B')
    assert tr.lines == ['B', 'c']



#! /usr/bin/env python
from setuptools import setup
import sys
import os

_PY3 = sys.version_info[0] == 3

def linelist(text):
    """
    Returns each non-blank line in text enclosed in a list.
    """
    return [ l.strip() for l in text.strip().splitlines() if l.split() ]

    # The double-mention of l.strip() is yet another fine example of why
    # Python needs en passant aliasing.

def getmetadata(filepath):
    """
    Return a dictionary of metadata from a file, without importing it. This
    trick needed because importing can set off ImportError, in that setup.py
    runs by definition before the modules it sets up (or their dependencies) are
    available.
    """
    if _PY3:
        exec(open(filepath).read())
        return vars()
    else:
        execfile(filepath)
        return locals()

metadata = getmetadata('./say/version.py')

setup(
    name='say',
    version=metadata['__version__'],
    author='Jonathan Eunice',
    author_email='jonathan.eunice@gmail.com',
    description='Super-simple templated printing. E.g.: say("Hello, {whoever}!", indent=1)',
    long_description=open('README.rst').read(),
    url='https://bitbucket.org/jeunice/say',
    packages=['say'],
    install_requires=['six>=1.4.1', 'options>=1.0.3', 'stuf>=0.9.12', 'simplere>=1.0', 'mementos>=1.0',
                      'ansicolors'],
    tests_require = ['tox', 'pytest', 'six'],
    zip_safe = True,
    keywords='print format template interpolate say',
    classifiers=linelist("""
        Development Status :: 4 - Beta
        Operating System :: OS Independent
        License :: OSI Approved :: BSD License
        Intended Audience :: Developers
        Programming Language :: Python
        Programming Language :: Python :: 2.6
        Programming Language :: Python :: 2.7
        Programming Language :: Python :: 3
        Programming Language :: Python :: 3.2
        Programming Language :: Python :: 3.3
        Programming Language :: Python :: Implementation :: CPython
        Programming Language :: Python :: Implementation :: PyPy
        Topic :: Software Development :: Libraries :: Python Modules
        Topic :: Printing
    """)
)



"""
A text object that uses say to aggregate lines, say in a file or script.
"""

from say.core import fmt
import inspect
import sys
import os
import re
from simplere import Re

_PY3 = sys.version_info[0] > 2
if _PY3:
    basestring = str


class Text(object):

    def __init__(self, data=None, interpolate=True, dedent=True):
        self._lines = []
        if data:
            lines = self._data_to_lines(data)
            callframe = inspect.currentframe().f_back
            self.extend(lines, callframe, interpolate, dedent)

    def _dedent_lines(self, lines):
        """
        Given a list of lines, remove a leading or trailing blank line (if any),
        as well as common leading blank prefixes.
        """
        if lines and lines[0].strip() == "":
            lines.pop(0)
        if lines and lines[-1].strip() == "":
            lines.pop()
        if lines:
            nonblanklines = [line for line in lines if line.strip() != ""]
            # piggyback os.path function to compute longest common prefix
            prefix = os.path.commonprefix(nonblanklines)
            # but then determine how much of that prefix is blank
            prelen, maxprelen = 0, len(prefix)
            while prelen < maxprelen and prefix[prelen] == ' ':
                prelen += 1
            if prelen:
                lines = [line[prelen:] for line in lines]
        return lines

    def __add__(self, data):
        """
        Add the contents of the self with the contents of the other item, returning
        a next text.
        """
        result = self.copy()
        lines = self._data_to_lines(data)
        callframe = inspect.currentframe().f_back
        result.extend(lines, callframe, interpolate=True, dedent=True)
        return result

    def __or__(self, data):
        """
        Add the contents of the self with the contents of the other item, returning
        a next text.
        """
        result = self.copy()
        lines = self._data_to_lines(data)
        callframe = inspect.currentframe().f_back
        result.extend(lines, callframe, interpolate=True, dedent=False)
        return result

    def __and__(self, data):
        """
        Add the contents of the self with the contents of the other item, returning
        a next text.
        """
        result = self.copy()
        lines = self._data_to_lines(data)
        result._lines.extend(lines)
        return result

    # TODO: Add the final 'yes dedent, no interpret' operation

    def __iadd__(self, data):
        """
        In-place add the text or lines contained in data, with auto-dedent.
        """
        lines = self._data_to_lines(data)
        callframe = inspect.currentframe().f_back
        self.extend(lines, callframe, interpolate=True, dedent=True)
        return self

    def __ior__(self, data):
        """
        In-place add the text or lines contained in data, with NO auto-dedent.
        """
        lines = self._data_to_lines(data)
        callframe = inspect.currentframe().f_back
        self.extend(lines, callframe, interpolate=True, dedent=False)
        return self

    def __iand__(self, data):
        """
        In-place add the text or lines contained in data, with NO auto-dedent
        and NO iterpolation.
        """
        lines = self._data_to_lines(data)
        self._lines.extend(lines)
        return self

    def __contains__(self, other):
        other = str(other)
        return other in self.text

    def append(self, line, callframe=None, interpolate=True):
        if interpolate:
            callframe = callframe or inspect.currentframe().f_back
            line = fmt(line, _callframe=callframe)
        self._lines.append(line)

    def extend(self, lines, callframe=None, interpolate=True, dedent=True):
        if dedent:
            lines = self._dedent_lines(lines)
        if interpolate:
            callframe = callframe or inspect.currentframe().f_back
        for line in lines:
            self.append(line, callframe, interpolate)

    # TODO: consider whether append and extend should be calls to insert

    def insert(self, i, data, callframe=None, interpolate=True, dedent=True):
        lines = self._data_to_lines(data)
        if dedent:
            lines = self._dedent_lines(lines)
        if interpolate:
            callframe = callframe or inspect.currentframe().f_back
            lines = [fmt(line, _callframe=callframe) for line in lines]
        for line in reversed(lines):
            self._lines.insert(i, line)

    def __getitem__(self, n):
        return self._lines[n]

    def __setitem__(self, n, value):
        self._lines[n] = value.rstrip('\n')

    def __len__(self):
        return len(self._lines)

    def __iter__(self):
        return iter(self._lines)

    def _data_to_lines(self, data):
        if isinstance(data, Text):
            return data.lines
        elif isinstance(data, list):
            return [line.rstrip('\n') for line in data]
        else:
            return data.splitlines()

    @property
    def text(self):
        return '\n'.join(self._lines)

    @text.setter
    def text(self, data):
        self._lines = self._data_to_lines(data)

    @property
    def lines(self):
        return self._lines[:]

    @lines.setter
    def lines(self, somelines):
        self._lines = [line.rstrip('\n') for line in somelines]

    def __str__(self):
        return self.text

    def __repr__(self):
        return 'Text({0}, {1} lines)'.format(id(self), len(self._lines))

    def copy(self):
        """
        Make a copy.
        """
        newt = Text()
        newt._lines = self._lines[:]
        return newt

    def replace(self, target, replacement=None):
        """
        Replace all instances of the target string with the replacement string.
        Works in situ, contra str.replace().
        """
        if target and isinstance(target, dict) and replacement is None:
            replacement = ''
            for (targ, repl) in target.items():
                repl = str(repl)
                for i, line in enumerate(self._lines):
                    self._lines[i] = line.replace(targ, repl)
                    replacement += repl  # for newline check below
        else:
            replacement = str(replacement)
            for i, line in enumerate(self._lines):
                self._lines[i] = line.replace(target, replacement)

        #  recalibrate lines, if new lines included in replacement
        if replacement and '\n' in replacement:
            self._lines = self._data_to_lines(self.text)

        return self

    # TODO: add doc note about caution re dict-based replacements; order is
    # base on dict iteration order, and subsequent can overwrite previous

    def re_replace(self, target, replacement):
        """
        Regular expression replacement. Target is either compiled re object or
        string that will be compiled into one. Replacement is either string or
        function that takes re match object as a parameter and returns replacement
        string.
        """
        target = Re(target) if not isinstance(target, Re) else target
        for i, line in enumerate(self._lines):
            self._lines[i] = target.sub(replacement, line)
        return self

    # TODO: for some reason, replacement not getting to be a simplere.ReMatch obj
    # TODO: extend dict-style invocation to re_replace

    def read_from(self, filepath, interpolate=True, dedent=True):
        """
        Reads lines from the designated file, appending them to the end of the
        given Text. By default, interpolates and dedents any {}
        expressions.
        """
        lines = open(filepath).read().splitlines()
        caller = inspect.currentframe().f_back
        self.extend(lines, caller, interpolate=interpolate, dedent=dedent)
        return self

    def write_to(self, filepath, append=False, encoding='utf-8'):
        mode = "a" if append else "w"
        with open(filepath, "w") as f:
            f.write(self.text.encode(encoding))

    # TODO: add a backup feature

# TODO: Consider additional `list` and `str` methods
# TODO: Possibly add in-place replacement, modification, etc.
# TODO: Extend tests



from say import *
import six

def test_basic_styling():
    fmt = Fmt()

    assert fmt('this', style='green+underline') == six.u('\x1b[32;4mthis\x1b[0m')
    assert fmt('this', style='bold+red') == six.u('\x1b[31;1mthis\x1b[0m')


def test_readme_example():
    fmt = Fmt()
    fmt.style(stars=lambda x: fmt('*** ', style='red') + \
                              fmt(x,      style='black') + \
                              fmt(' ***', style='red'))

    message = 'Warning, problem!'
    assert fmt(message, style='stars') == six.u('\x1b[31m*** \x1b[0m\x1b[30mWarning, problem!\x1b[0m\x1b[31m ***\x1b[0m')


def test_readme_example2():
    fmt = Fmt()
    name = 'Joe'
    assert six.u('His name is ') + fmt(name, style='blue+underline') == six.u('His name is \x1b[34;4mJoe\x1b[0m')
    assert fmt('His name is {name:style=blue+underline}') == six.u('His name is \x1b[34;4mJoe\x1b[0m')
    assert fmt('His name is {name:style="blue+underline"}') == six.u('His name is \x1b[34;4mJoe\x1b[0m')
    assert fmt("His name is {name:style='blue+underline'}") == six.u('His name is \x1b[34;4mJoe\x1b[0m')


def test_wrapping_example():
    fmt = Fmt()
    text = "Move over, Coke. It looks like Apple is the real thing. The tech giant has ended Coca-Cola's 13-year run as the world's most valuable brand on a highly regarded annual list."

    fmt = Fmt()
    fmt.set(prefix=numberer(template=color('{n:>3}: ', fg='blue')), \
            wrap=40)
    assert fmt(text) == six.u("\x1b[34m  1: \x1b[0mMove over, Coke. It looks\n\x1b[34m  2: \x1b[0mlike Apple is the real\n\x1b[34m  3: \x1b[0mthing. The tech giant has\n\x1b[34m  4: \x1b[0mended Coca-Cola's 13-year\n\x1b[34m  5: \x1b[0mrun as the world's most\n\x1b[34m  6: \x1b[0mvaluable brand on a highly\n\x1b[34m  7: \x1b[0mregarded annual list.")

    # now reset so numbering starts back at 1
    fmt = Fmt()
    fmt.set(prefix=numberer(template=color('{n:>3}: ', fg='blue')), \
            wrap=40)
    assert fmt(text, style='red') == six.u("\x1b[34m  1: \x1b[0m\x1b[31mMove over, Coke. It looks\x1b[0m\n\x1b[34m  2: \x1b[0m\x1b[31mlike Apple is the real\x1b[0m\n\x1b[34m  3: \x1b[0m\x1b[31mthing. The tech giant has\x1b[0m\n\x1b[34m  4: \x1b[0m\x1b[31mended Coca-Cola's 13-year\x1b[0m\n\x1b[34m  5: \x1b[0m\x1b[31mrun as the world's most\x1b[0m\n\x1b[34m  6: \x1b[0m\x1b[31mvaluable brand on a highly\x1b[0m\n\x1b[34m  7: \x1b[0m\x1b[31mregarded annual list.\x1b[0m")

def test_color():
    assert color('text', fg='green') == '\x1b[32mtext\x1b[0m'
    assert color('more', fg='blue', bg='yellow') == '\x1b[34;43mmore\x1b[0m'

def test_styled():
    assert color('test', fg='green', style='bold') == styled('test', 'green+bold')
    assert color('test', fg='green', style='bold') == styled('test', 'bold+green')
    assert color('test', fg='green', bg='red', style='bold') == styled('test', 'green+red+bold')
    assert color('test', fg='green', bg='red', style='bold') == styled('test', 'bold+green+red')
    assert color('test', fg='green', bg='red', style='bold') == styled('test', 'bold|green|red')
    assert color('test', fg='green', bg='red', style='bold') == styled('test', 'bold,green,red')


def test_in_or_out():
    fmt = Fmt()

    x = 12
    assert fmt(x, style='blue+white+underline') == fmt("{x:style=blue+white+underline}")

    fmt.style(bwu=autostyle('blue+white+underline'))
    fmt.style(bwu2='blue+white+underline')

    assert fmt(x, style='bwu') == fmt(x, style='blue+white+underline')
    assert fmt(x, style='bwu') == fmt(x, style='bwu2')



"""
Home for separable utility functions used in say
"""

import itertools
import sys
import six
import types
import csv
import io
if six.PY3:
    from codecs import getencoder
    unicode = str
    from io import StringIO
else:
    from StringIO import StringIO

def is_string(v):
    """
    Is the value v a string? Useful especially in making a test that works on
    both Python 2.x and Python 3.x
    """
    return isinstance(v, six.string_types)

def opened(f):
    """
    If f is a string, consider it a file name and return it, ready for writing.
    Else, assume it's an open file. Just return it.
    """
    if isinstance(f, list):
        return [ opened(ff) for ff in f ]
    else:
        return io.open(f, "w") if is_string(f) else f
        # NB use io.open not plain open to get auto-encoding to UTF-8 in Python 2

def encoded(u, encoding):
    """
    Encode string u (denoting it is expected to be in Unicode) if there's
    encoding to be done. Tries to mask the difference between Python 2 and 3,
    which have different models of string processing, and different codec APIs
    and quirks. Some Python 3 encoders further require ``bytes`` in, not
    ``str``. These are first encoded into utf-8, encoded, then decoded.
    """
    if not encoding:
        return u
    elif six.PY3:
        try:
            return getencoder(encoding)(u)[0]
        except LookupError:
            name = encoding.replace('-', '')  # base-64 becomes base64 e.g.
            bytesout = getencoder(name + '_codec')(u.encode('utf-8'))[0]
            if name in set(['base64', 'hex', 'quopri', 'uu']):
                return bytesout.decode('utf-8')
            else:
                return bytesout

        # NB PY3 requires lower-level interface for many codecs. s.encode('utf-8')
        # works fine, but others do not. Some codecs convert bytes to bytes,
        # and are not properly looked up by nickname (e.g. 'base64'). These are
        # managed by first encoding into utf-8, then if it makes sense decoding
        # back into a string. The others are things like bz2 and zlib--binary
        # encodings that have little use to us here.

        # There are also be some slight variations in results that will make
        # testing more fun. Possibly related to adding or not adding a terminal
        # newline.
    else:
        # Python 2 may not have the best handling of Unicode, but by
        # by God its encode operations are straightforward!
        return u.encode(encoding)


def flatten(*args):
    """
    Like itertools.chain(), but will pretend that single scalar values are
    singleton lists. Convenient for iterating over values whether they're lists
    or singletons.
    """
    flattened = [x if isinstance(x, (list, tuple)) else [x] for x in args]
    return itertools.chain(*flattened)

    # would use ``hasattr(x, '__iter__')`` rather than ``isinstance(x, (list, tuple))``,
    # but other objects like file have ``__iter__``, which screws things up


def next_str(g):
    """
    Given a generator g, return its next result as a unicode string. If not a
    generator, just return the value as a unicode string.
    """
    try:
        value = next(g)
    except TypeError:
        value = g if g is not None else ''
    return unicode(value)

